import numpy as np

from flex.flex import FlexFile


def test_read_write(tmp_fname):
    file = FlexFile()
    file.header["Hello"] = "World"
    file.write(tmp_fname)

    del file
    f2 = FlexFile.read(tmp_fname)

    assert isinstance(f2.header, dict)
    assert f2.header["Hello"] == "World"


def test_header_numpy_arrays(tmp_fname):
    file = FlexFile()
    file.header["Hello"] = np.ones(3)
    file.write(tmp_fname)

    del file
    f2 = FlexFile.read(tmp_fname)

    assert isinstance(f2.header, dict)
    assert np.all(np.asarray(f2.header["Hello"]) == 1)



import json
import sys
from json.encoder import (
    INFINITY,
    _make_iterencode,
    c_make_encoder,
    encode_basestring,
    encode_basestring_ascii,
)
import importlib

import numpy as np
from astropy import units, coordinates, time
import datetime

INFINITY_VALUE = sys.float_info.max


class FlexJSONEncoder(json.JSONEncoder):
    """
    Custom JSON Encoder for Flex files
    this overrides the float behaviour to default to null for nan values
    and the maximum floating point value for +/-infinity
    It also overrides the default method to handle numpy values and cast them 
    to their python base type
    """

    def iterencode(self, obj, _one_shot=False):
        """Encode the given object and yield each string
        representation as available.

        For example::

            for chunk in JSONEncoder().iterencode(bigobject):
                mysocket.write(chunk)

        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = encode_basestring_ascii
        else:
            _encoder = encode_basestring

        def floatstr(
            obj,
            allow_nan=self.allow_nan,
            _repr=float.__repr__,
            _inf=INFINITY,
            _neginf=-INFINITY,
        ):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if obj != obj:  # obj is NaN
                text = "null"
            elif obj == _inf:
                text = _repr(INFINITY_VALUE)
            elif obj == _neginf:
                text = _repr(-INFINITY_VALUE)
            else:
                return _repr(obj)

            return text

        _iterencode = _make_iterencode(
            markers,
            self.default,
            _encoder,
            self.indent,
            floatstr,
            self.key_separator,
            self.item_separator,
            self.sort_keys,
            self.skipkeys,
            _one_shot,
        )
        return _iterencode(obj, 0)

    def default(self, obj):
        if hasattr(obj, "to_dict"):
            data = obj.to_dict()
            data["__module__"] = obj.__class__.__module__
            data["__class__"] = obj.__class__.__name__
            return data
        if isinstance(obj, coordinates.SkyCoord):
            return {
                "__module__": obj.__class__.__module__,
                "__class__": obj.__class__.__name__,
                "frame": obj.frame.name,
                "ra": self.default(obj.ra),
                "dec": self.default(obj.dec),
            }
        if isinstance(obj, coordinates.earth.EarthLocation):
            return {
                "__module__": obj.__class__.__module__,
                "__class__": obj.__class__.__name__,
                "x": self.default(obj.x),
                "y": self.default(obj.y),
                "z": self.default(obj.z),
            }
        if isinstance(obj, units.UnitBase):
            return {
                "__module__": "astropy.units",
                "__class__": "Unit",  # Unit can parse everything no problem
                "value": str(obj),
            }
        if isinstance(obj, units.DexUnit):
            return {
                "__module__": "astropy.units",
                "__class__": "Unit",
                "value": str(obj),
            }
        if isinstance(obj, units.quantity.Quantity):
            return {
                "__module__": obj.__class__.__module__,
                "__class__": obj.__class__.__name__,
                "value": obj.value,
                "unit": str(obj.unit),
            }
        if isinstance(obj, time.Time):
            return {
                "__module__": obj.__class__.__module__,
                "__class__": obj.__class__.__name__,
                "format": obj.format,
                "value": self.default(obj.value),
            }
        if isinstance(obj, datetime.datetime):
            return {
                "__module__": obj.__class__.__module__,
                "__class__": obj.__class__.__name__,
                "tzinfo": obj.tzinfo,
                "year": obj.year,
                "month": obj.month,
                "day": obj.day,
                "hour": obj.hour,
                "minute": obj.minute,
                "second": obj.second,
                "microsecond": obj.microsecond,
            }

        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.str):
            return str(obj)
        if isinstance(obj, bytes):
            return obj.decode()

        return super().default(obj)


class FlexJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        kwargs["object_hook"] = self._object_hook
        super().__init__(*args, **kwargs)

    def _object_hook(self, obj):
        # If we specified a class and module use this:
        # but its not the lowest level of the header!
        if (
            isinstance(obj, dict)
            and "__class__" in obj
            and "__module__" in obj
            and not obj.get("__header__", False)
        ):
            module = importlib.import_module(obj["__module__"])
            cls = getattr(module, obj["__class__"])

            if obj["__class__"] == "Quantity" and obj["value"] is None:
                obj["value"] = float("NaN")

            for k, v in obj.items():
                if k not in ["__class__", "__module__"]:
                    obj[k] = self._object_hook(v)

            args = (obj["value"],) if "value" in obj else ()
            exceptions = ["__class__", "__module__", "value"]
            kwargs = {k: v for k, v in obj.items() if k not in exceptions}

            if hasattr(cls, "from_json"):
                return cls.from_json(*args, **kwargs)

            return cls(*args, **kwargs)
        # Otherwise just return it
        return obj



from io import BytesIO, TextIOWrapper

import pandas as pd

from ..flex import FlexExtension

try:
    from astropy.io import fits
    from astropy.table import Table
except ImportError:
    fits = Table = None


class TableExtension(FlexExtension):
    data_extension = "parquet"

    def __init__(self, header={}, data=None, cls=None):
        super().__init__(header=header, cls=cls)
        self.data = data

    @classmethod
    def _prepare_table(cls, name: str, data: pd.DataFrame):
        bio = BytesIO()
        data.to_parquet(bio, index=True)
        info = cls._get_tarinfo_from_bytesio(name, bio)
        return info, bio

    def _prepare(self, name: str):
        cls = self.__class__
        header_fname = f"{name}/header.json"
        data_fname = f"{name}/data.{cls.data_extension}"
        header_info, header_bio = cls._prepare_json(header_fname, self.header)
        data_info, data_bio = cls._prepare_table(data_fname, self.data)

        return [(header_info, header_bio), (data_info, data_bio)]

    @classmethod
    def _parse_table(cls, bio: BytesIO):
        b = BytesIO(bio.read())
        data = pd.read_parquet(b)
        return data

    @classmethod
    def _parse(cls, header: dict, members: list):
        bio = members[f"data.{cls.data_extension}"]
        data = cls._parse_table(bio)
        ext = cls(header=header, data=data)
        return ext

    def to_dict(self):
        obj = {"header": self.header, "data": self.data.to_dict(orient="records")}
        return obj

    @classmethod
    def from_dict(cls, header: dict, data: dict):
        data = pd.DataFrame.from_records(data["data"])
        obj = cls(header, data)
        return obj

    def to_fits(self):
        header = self._prepare_fits_header(self.header)
        table = Table.from_pandas(self.data)
        hdu = fits.BinTableHDU(table, header)
        return hdu

    @classmethod
    def from_fits(cls, header, data):
        df = Table(data).to_pandas()
        obj = cls(header, df)
        return obj


class AsciiTableExtension(TableExtension):
    data_extension = "txt"

    @classmethod
    def _prepare_table(cls, name: str, data: pd.DataFrame):
        tio = TextIOWrapper(BytesIO(), "utf-8")
        data.to_csv(tio, index=False)
        bio = tio.detach()
        info = cls._get_tarinfo_from_bytesio(name, bio)
        return info, bio

    @staticmethod
    def _parse_table(bio: BytesIO):
        data = pd.read_csv(bio)
        return data


class JSONTableExtension(TableExtension):
    data_extension = "json"

    @classmethod
    def _prepare_table(cls, name: str, data: pd.DataFrame):
        tio = TextIOWrapper(BytesIO(), "utf-8")
        data.to_json(tio, orient="records")
        bio = tio.detach()
        info = cls._get_tarinfo_from_bytesio(name, bio)
        return info, bio

    @staticmethod
    def _parse_table(bio: BytesIO):
        data = pd.read_json(bio, orient="records")
        return data



import numpy as np
import pandas as pd

from .flex import FlexFile
from .extensions.bindata import BinaryDataExtension
from .extensions.tabledata import TableExtension


def write(filename, **data):
    header = {}
    extensions = {}
    for key, value in data.items():
        if hasattr(value, "__flex_save__"):
            extensions[key] = value
        elif type(value) is np.ndarray:
            extensions[key] = BinaryDataExtension(data=value)
        elif isinstance(value, pd.DataFrame):
            extensions[key] = TableExtension(data=value)
        else:
            header[key] = value

    ff = FlexFile(header=header, extensions=extensions)
    ff.write(filename)


def read(filename):
    return FlexFile.read(filename)



import ast
import importlib
import json
import tarfile
from io import BytesIO, TextIOWrapper
from tarfile import TarInfo
import warnings

import numpy as np
from numpy.core.defchararray import startswith

from . import __version__
from .json_encoder import FlexJSONDecoder, FlexJSONEncoder

try:
    from astropy.io import fits
except ImportError:
    fits = None

FITS_DTYPES = (
    str,
    int,
    float,
    complex,
    bool,
    np.floating,
    np.integer,
    np.complexfloating,
    np.bool_,
)


class FlexBase:
    @classmethod
    def _prepare_json(cls, fname: str, data: dict):
        tio = TextIOWrapper(BytesIO(), "utf-8")
        kwargs = dict(cls=FlexJSONEncoder, allow_nan=False, skipkeys=False)
        json.dump(data, tio, **kwargs)
        bio = tio.detach()
        info = cls._get_tarinfo_from_bytesio(fname, bio)
        return info, bio

    @staticmethod
    def _parse_json(bio):
        data = json.load(bio, cls=FlexJSONDecoder)
        return data

    @classmethod
    def _read_json(cls, file, name):
        bio = file.extractfile(name)
        data = cls._parse_json(bio)
        return data

    @staticmethod
    def _get_tarinfo_from_bytesio(fname, bio):
        info = TarInfo(fname)
        info.size = bio.tell()
        bio.seek(0)
        return info

    def _prepare_fits_header(self, header):
        def floatstr(obj):
            if isinstance(obj, FITS_DTYPES):
                return obj
            if isinstance(obj, (str, np.str)):
                return ascii(obj)[1:-1]
            return ascii(repr(obj))[1:-1]

        def check_length(k, v):
            if (
                not isinstance(v, str)
                or not k.startswith("HIERARCH ")
                or len(k) + len(v) + 5 <= 80
            ):
                return v
            else:
                warnings.warn(
                    "The value of keyword %s is too long, it has been truncated"
                    "To avoid this reduce the length of the keyword to less than 8 characters"
                    % k
                )
                return ascii(v[: 80 - len(k) - 6])[1:-1]

        # FITS only allows strings and integers, so convert everything into str
        # But only in ASCII representation
        header = {k: floatstr(v) for k, v in self.header.items()}
        # The key must be 8 letters or less, otherwise use the HIERARCH extension
        # usually this is handled by the fits package, but here we need to check
        # the length of the key + value, see next point
        header = {k if len(k) <= 8 else "HIERARCH " + k: v for k, v in header.items()}
        # Also the length of the key + value is limited to 80 characters (including the delimiter '=')
        # If the key is not a HIERARCH extensions, we can use the CONTINUE extension to make it work
        # But HIERARCH and CONTINUE can not be combined so we truncate the value if it is a HIERARCH keyword
        header = {k: check_length(k, v) for k, v in header.items()}
        header = fits.Header(header)
        return header

    @classmethod
    def _parse_fits_header(cls, header):
        INFINITY_VALUE = float("inf")
        NaN = float("NaN")

        def floatstr(obj):
            if obj == "nan":
                return NaN
            if obj == "inf":
                return INFINITY_VALUE
            if obj == "-inf":
                return -INFINITY_VALUE
            if isinstance(obj, str):
                if obj.startswith("array"):
                    return np.array(ast.literal_eval(obj[6:-1]))
                try:
                    return ast.literal_eval(obj)
                except:
                    pass
            return obj

        header = {k: floatstr(v) for k, v in header.items()}
        return header


class FlexExtension(FlexBase):
    def __init__(self, header={}, cls=None, **kwargs):
        self.header = header
        if cls is not None:
            (self.header["__module__"], self.header["__class__"],) = cls.rsplit(".", 1)
        else:
            self.header["__module__"] = self.__class__.__module__
            self.header["__class__"] = self.__class__.__name__
        self.header["__header__"] = True

    def _prepare(self, name: str):
        raise NotImplementedError

    @classmethod
    def _parse(cls, header: dict, members: dict):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError

    @classmethod
    def from_dict(cls, header: dict, data):
        raise NotImplementedError


class JsonDataExtension(FlexExtension):
    def __init__(self, header={}, data={}, cls=None):
        super().__init__(header=header, cls=cls)
        self.data = data

    def _prepare(self, name: str):
        cls = self.__class__
        header_fname = f"{name}/header.json"
        data_fname = f"{name}/data.json"
        header_info, header_bio = cls._prepare_json(header_fname, self.header)
        data_info, data_bio = cls._prepare_json(data_fname, self.data)

        return [(header_info, header_bio), (data_info, data_bio)]

    @classmethod
    def _parse(cls, header: dict, members: dict):
        bio = members["data.json"]
        data = cls._parse_json(bio)
        ext = cls(header=header, data=data)
        return ext

    def to_dict(self):
        obj = {"header": self.header, "data": self.data}
        return obj

    @classmethod
    def from_dict(cls, header: dict, data: dict):
        arr = data["data"]
        obj = cls(header, arr)
        return obj


class FlexFile(FlexBase):
    def __init__(self, header={}, extensions={}, fileobj=None):
        if isinstance(header, str):
            other = self.read(header)
            self.header = other.header
            self.extensions = other.extensions
            self.fileobj = other.fileobj
        else:
            self.header = header
            self.extensions = extensions
            self.fileobj = fileobj
            self.header["__version__"] = __version__
            self.header["__header__"] = True

    def __getitem__(self, key):
        return self.extensions[key]

    def __setitem__(self, key, value):
        self.extensions[key] = value

    def write(self, fname: str, compression=False):
        # Write the header
        cls = self.__class__
        # Update header with current metadata
        self.header["__version__"] = __version__
        self.header["__header__"] = True

        info, bio = cls._prepare_json("header.json", self.header)
        extensions = []
        for name, ext in self.extensions.items():
            if isinstance(ext, FlexExtension):
                extensions += ext._prepare(name)
            elif hasattr(ext, "__flex_save__"):
                extensions += ext.__flex_save__()._prepare(name)
            elif hasattr(ext, "to_dict"):
                extensions += JsonDataExtension(
                    data=ext.to_dict(),
                    cls=f"{ext.__class__.__module__}.{ext.__class__.__name__}",
                )
            else:
                raise ValueError(f"Could not save extension {name}")

        mode = "w:" if not compression else "w:gz"
        with tarfile.open(fname, mode) as file:
            file.addfile(info, bio)
            for ext in extensions:
                file.addfile(ext[0], ext[1])

    @classmethod
    def _read_ext_class(cls, ext_header):
        try:
            ext_module = ext_header["__module__"]
            ext_class = ext_header["__class__"]
        except KeyError:
            ext_module = ext_header["extension_module"]
            ext_class = ext_header["extension_class"]

        ext_module = importlib.import_module(ext_module)
        ext_class = getattr(ext_module, ext_class)
        return ext_class

    @classmethod
    def read(cls, fname: str, mmap=False):
        handle = open(fname, "rb")
        # If we allow mmap.ACCESS_WRITE, we invalidate the checksum
        # So I think the best solution is to use COPY
        # This also prevents the user accidentially messing up the files
        if mmap:
            mapped = mmap.mmap(handle.fileno(), 0, access=mmap.ACCESS_COPY)
            file = tarfile.open(mode="r", fileobj=mapped)
        else:
            file = tarfile.open(fname, mode="r")

        header = cls._read_json(file, "header.json")

        names = file.getnames()
        names = np.array([n for n in names if n != "header.json"])
        if len(names) != 0:
            # If the file was created using Windows style paths
            ext = np.char.replace(names, "\\", "/")
            ext = np.char.partition(ext, "/")
            ext = ext[:, 0]
            ext, mapping = np.unique(ext, return_inverse=True)
        else:
            ext, mapping = [], None

        extensions = {}
        for i, name in enumerate(ext):
            # Determine the files contributing to this extension
            members = names[mapping == i]
            ext_header = ""
            ext_other = []
            for n in members:
                if n.endswith("header.json"):
                    ext_header = n
                else:
                    ext_other += [n]

            # Determine the extension class and module
            ext_header = cls._read_json(file, ext_header)
            ext_class = cls._read_ext_class(ext_header)

            # TODO: lazy load the extensions?
            ext_other = {
                other[len(name) + 1 :]: file.extractfile(other) for other in ext_other
            }
            if issubclass(ext_class, FlexExtension):
                exten = ext_class._parse(ext_header, ext_other)
            elif hasattr(ext_class, "__flex_load__"):
                exten = ext_class.__flex_load__(ext_header, ext_other)
            elif hasattr(ext_class, "from_dict"):
                temp_exten = JsonDataExtension._parse(ext_header, ext_other)
                exten = ext_class.from_dict(temp_exten.data)
            else:
                raise ValueError(f"Could not decode extension {name}")

            extensions[name] = exten

        return cls(header=header, extensions=extensions, fileobj=file)

    def close(self):
        if self.fileobj is not None:
            self.fileobj.fileobj.close()
            self.fileobj.close()

    def to_dict(self):
        obj = {"header": self.header}
        for name, ext in self.extensions.items():
            obj[name] = ext.to_dict()
        return obj

    @classmethod
    def from_dict(cls, data: dict):
        header = {}
        extensions = {}
        for name, ext in data.items():
            if name == "header":
                header = ext
                continue

            ext_header = ext["header"]
            ext_class = cls._read_ext_class(ext_header)
            del ext["header"]
            exten = ext_class.from_dict(ext_header, ext)
            extensions[name] = exten

        obj = cls(header, extensions)
        return obj

    def to_json(self, fp=None):
        obj = self.to_dict()
        kwargs = dict(cls=FlexJSONEncoder, allow_nan=False, skipkeys=False)
        if fp is None:
            obj = json.dumps(obj, **kwargs)
            return obj
        elif isinstance(fp, str):
            with open(fp, "w") as f:
                json.dump(obj, f, **kwargs)
        else:
            json.dump(obj, fp, **kwargs)

    @classmethod
    def from_json(cls, obj):
        try:
            obj = json.loads(obj, cls=FlexJSONDecoder)
        except json.decoder.JSONDecodeError as ex:
            # Its already a json string
            with open(obj, "r") as f:
                obj = json.load(f, cls=FlexJSONDecoder)
        obj = cls.from_dict(obj)
        return obj

    def to_fits(self, fname=None, overwrite=False):
        header = self._prepare_fits_header(self.header)
        primary = fits.PrimaryHDU(header=header)
        hdus = [primary]
        for ext_name, ext_data in self.extensions.items():
            ext_hdu = ext_data.to_fits()
            ext_hdu.header["EXTNAME"] = ext_name
            hdus += [ext_hdu]

        hdulist = fits.HDUList(hdus)
        if fname is not None:
            hdulist.writeto(fname, overwrite=overwrite)
        return hdulist

    @classmethod
    def from_fits(cls, fname):
        hdulist = fits.open(fname)
        header = cls._parse_fits_header(hdulist[0].header)
        extensions = {}
        for i in range(1, len(hdulist)):
            hdu = hdulist[i]
            ext_header = cls._parse_fits_header(hdu.header)
            ext_data = hdu.data
            ext_name = ext_header["EXTNAME"]
            ext_class = cls._read_ext_class(ext_header)
            extension = ext_class.from_fits(ext_header, ext_data)
            extensions[ext_name] = extension

        obj = cls(header, extensions)
        return obj



import numpy as np

from flex.flex import FlexFile
from flex.extensions.bindata import BinaryDataExtension, MultipleDataExtension


def test_read_write(tmp_fname):
    file = FlexFile()
    ext = BinaryDataExtension(data=np.zeros(10))
    file.extensions["img"] = ext

    file.write(tmp_fname)

    del file
    f2 = FlexFile.read(tmp_fname)

    assert f2["img"].data.size == 10
    assert np.all(f2["img"].data == 0)


def test_empty_data(tmp_fname):
    file = FlexFile()
    ext = BinaryDataExtension(data=[])
    file.extensions["img"] = ext

    file.write(tmp_fname)

    del file
    f2 = FlexFile.read(tmp_fname)

    assert f2["img"].data.size == 0


def test_empty_data_json(tmp_fname):
    file = FlexFile()
    ext = BinaryDataExtension(data=[])
    file.extensions["img"] = ext

    file.to_json(tmp_fname)

    del file
    f2 = FlexFile.from_json(tmp_fname)

    assert f2["img"].data.size == 0


def test_json(tmp_fname):
    file = FlexFile()
    ext = BinaryDataExtension(data=np.zeros(10))
    file.extensions["img"] = ext

    file.to_json(tmp_fname)

    del file
    f2 = FlexFile.from_json(tmp_fname)

    assert f2["img"].data.size == 10
    assert np.all(f2["img"].data == 0)


def test_read_write_multi(tmp_fname):
    file = FlexFile()
    ext = MultipleDataExtension(data={"bla": np.zeros(10)})
    file.extensions["img"] = ext

    file.write(tmp_fname)

    del file
    f2 = FlexFile.read(tmp_fname)

    assert f2["img"].data["bla"].size == 10
    assert np.all(f2["img"].data["bla"] == 0)


def test_json_multi(tmp_fname):
    file = FlexFile()
    ext = MultipleDataExtension(data={"bla": np.zeros(10)})
    file.extensions["img"] = ext

    file.to_json(tmp_fname)

    del file
    f2 = FlexFile.from_json(tmp_fname)

    assert f2["img"].data["bla"].size == 10
    assert np.all(f2["img"].data["bla"] == 0)



import pytest
import numpy as np
import pandas as pd

from flex.flex import FlexFile
from flex.extensions.tabledata import (
    TableExtension,
    AsciiTableExtension,
    JSONTableExtension,
)


@pytest.fixture
def table():
    df = pd.DataFrame(np.ones((10, 2)), columns=["A", "B"])
    return df


@pytest.fixture(
    params=[TableExtension, AsciiTableExtension, JSONTableExtension],
    ids=["table", "ascii", "json"],
)
def tableclass(request):
    return request.param


def test_read_write(tableclass, tmp_fname, table):
    file = FlexFile()
    ext = tableclass(data=table)
    file.extensions["tab"] = ext

    file.write(tmp_fname)

    del file
    f2 = FlexFile.read(tmp_fname)

    assert f2["tab"].data.size == 10 * 2
    assert np.all(f2["tab"].data == 1)
    assert len(f2["tab"].data.columns) == 2
    assert "A" in f2["tab"].data.columns
    assert "B" in f2["tab"].data.columns


def test_json(tableclass, tmp_fname, table):
    file = FlexFile()
    ext = tableclass(data=table)
    file.extensions["tab"] = ext

    file.to_json(tmp_fname)

    del file
    f2 = FlexFile.from_json(tmp_fname)

    assert f2["tab"].data.size == 10 * 2
    assert np.all(f2["tab"].data == 1)
    assert len(f2["tab"].data.columns) == 2
    assert "A" in f2["tab"].data.columns
    assert "B" in f2["tab"].data.columns



from setuptools import setup, find_packages
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="flex-format",
    url="https://github.com/AWehrhahn/flex-format",
    author="Ansgar Wehrhahn",
    author_email="ansgar.wehrhahn@physics.uu.se",
    description="A flexible file format for astronomical (and other) data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import base64
import mmap
from io import BytesIO
from tarfile import TarInfo

import numpy as np
from numpy.lib.format import _check_version, _read_array_header, read_magic

from ..flex import FlexExtension

try:
    from astropy.io import fits
    from astropy.io.fits.column import NUMPY2FITS
except ImportError:
    fits = None


class BinaryDataExtension(FlexExtension):
    def __init__(self, header={}, data=[], cls=None):
        super().__init__(header=header, cls=cls)
        self.data = np.asarray(data)

    @classmethod
    def _prepare_npy(cls, fname: str, data: np.ndarray):
        bio = BytesIO()
        np.save(bio, data)
        info = cls._get_tarinfo_from_bytesio(fname, bio)
        return info, bio

    def _prepare(self, name: str):
        cls = self.__class__
        header_fname = f"{name}/header.json"
        data_fname = f"{name}/data.npy"
        header_info, header_bio = cls._prepare_json(header_fname, self.header)
        data_info, data_bio = cls._prepare_npy(data_fname, self.data)

        return [(header_info, header_bio), (data_info, data_bio)]

    @staticmethod
    def _parse_npy(bio):
        mmapfile = bio.raw.fileobj
        if isinstance(mmapfile, mmap.mmap):
            version = read_magic(bio)
            _check_version(version)

            shape, fortran_order, dtype = _read_array_header(bio, version)
            if dtype.hasobject:
                msg = "Array can't be memory-mapped: Python objects in dtype."
                raise ValueError(msg)
            order = "F" if fortran_order else "C"
            offset = bio.tell()
            # Add the offset from the Wrapper file
            offset += bio.raw.offset
            data = np.ndarray.__new__(
                np.memmap,
                shape,
                dtype=dtype,
                buffer=mmapfile,
                offset=offset,
                order=order,
            )
            data._mmap = mmapfile
            data.offset = offset
            data.mode = "r+"
        else:
            b = BytesIO(bio.read())
            data = np.load(b)

        return data

    @classmethod
    def _parse(cls, header: dict, members: dict):
        bio = members["data.npy"]
        data = cls._parse_npy(bio)
        ext = cls(header=header, data=data)
        return ext

    @staticmethod
    def _np_to_dict(data):
        encoded = base64.b64encode(data.tobytes())
        encoded = encoded.decode("utf-8")
        obj = {
            "dtype": data.dtype.str,
            "shape": data.shape,
            "order": "C",
            "data": encoded,
        }
        return obj

    @staticmethod
    def _np_from_dict(data):
        decoded = data["data"].encode("utf-8")
        decoded = base64.b64decode(decoded)
        arr = np.frombuffer(decoded, dtype=data["dtype"])
        arr = arr.reshape(data["shape"])
        arr = np.require(arr, requirements=["W"])
        return arr

    def to_dict(self):
        cls = self.__class__
        obj = {
            "header": self.header,
            "data": cls._np_to_dict(self.data),
        }
        return obj

    @classmethod
    def from_dict(cls, header: dict, data: dict):
        arr = cls._np_from_dict(data["data"])
        obj = cls(header, arr)
        return obj

    @classmethod
    def from_json(cls, header=None, **data):
        return cls.from_dict(header, data)

    @staticmethod
    def _prepare_fits_array(value):
        try:
            fits_format = f"{value.dtype.kind}{value.dtype.itemsize}"
            fits_format = NUMPY2FITS[fits_format]
            fits_dim = None
            shape = value.shape
            if len(shape) > 1:
                size = np.prod(value.shape[1:])
                fits_format = "%i%s" % (size, fits_format)
                if len(shape) > 2:
                    fits_dim = str(value.shape[1:][::-1])
        except KeyError:
            fits_format = "D"
            value = value.astype("f8").ravel()
            fits_dim = None
        return fits_format, fits_dim, value

    def to_fits(self):
        cls = self.__class__
        header = self._prepare_fits_header(self.header)
        fits_format, fits_dim, value = cls._prepare_fits_array(self.data)
        column = fits.Column(name="data", format=fits_format, dim=fits_dim, array=value)
        hdu = fits.BinTableHDU.from_columns([column], header)
        return hdu

    @classmethod
    def from_fits(cls, header, data):
        arr = data["data"]
        obj = cls(header, arr)
        return obj


class MultipleDataExtension(BinaryDataExtension):
    def __init__(self, header={}, data={}, cls=None):
        super().__init__(header=header, cls=cls)
        self.data = dict(data)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def _prepare(self, name: str):
        cls = self.__class__

        header_fname = f"{name}/header.json"
        header_info, header_bio = cls._prepare_json(header_fname, self.header)
        result = [(header_info, header_bio)]

        for key, value in self.data.items():
            data_fname = f"{name}/{key}.npy"
            data_info, data_bio = cls._prepare_npy(data_fname, value)
            result += [(data_info, data_bio)]

        return result

    @classmethod
    def _parse(cls, header: dict, members: dict):
        data = {key[:-4]: cls._parse_npy(bio) for key, bio in members.items()}
        ext = cls(header=header, data=data)
        return ext

    def to_dict(self):
        cls = self.__class__
        obj = {"header": self.header}
        for name, data in self.data.items():
            obj[name] = cls._np_to_dict(data)
        return obj

    @classmethod
    def from_dict(cls, header: dict, data: dict):
        data = {name: cls._np_from_dict(d) for name, d in data.items()}
        obj = cls(header, data=data)
        return obj

    def to_fits(self):
        cls = self.__class__
        header = self._prepare_fits_header(self.header)
        columns = []
        for key, value in self.data.items():
            fits_format, fits_dim, value = cls._prepare_fits_array(value)
            # header["__shape_%s__" % key] = str(value.shape)
            column = fits.Column(
                name=key, format=fits_format, dim=fits_dim, array=value
            )
            columns += [column]
        hdu = fits.BinTableHDU.from_columns(columns, header)
        return hdu

    @classmethod
    def from_fits(cls, header, data):
        arr = {name: data[name] for name in data.names}
        obj = cls(header, arr)
        return obj
